@@ -85,6 +85,8 @@ int hidden sepol_set_policydb(policydb_t * p)
 int sepol_set_policydb_from_file(FILE * fp)
 {
 	struct policy_file pf;
+
+	policy_file_init(&pf);
 	pf.fp = fp;
 	pf.type = PF_USE_STDIO;
 	if (mypolicydb.policy_type)
@@ -927,6 +929,58 @@ static int convert_context(sepol_security_id_t key __attribute__ ((unused)),
 	return rc;
 }
 
+/* Reading from a policy "file". */
+int hidden next_entry(void *buf, struct policy_file *fp, size_t bytes)
+{
+	size_t nread;
+
+	switch (fp->type) {
+	case PF_USE_STDIO:
+		nread = fread(buf, bytes, 1, fp->fp);
+
+		if (nread != 1)
+			return -1;
+		break;
+	case PF_USE_MEMORY:
+		if (bytes > fp->len)
+			return -1;
+		memcpy(buf, fp->data, bytes);
+		fp->data += bytes;
+		fp->len -= bytes;
+		break;
+	default:
+		return -1;
+	}
+	return 0;
+}
+
+size_t hidden put_entry(const void *ptr, size_t size, size_t n,
+			struct policy_file *fp)
+{
+	size_t bytes = size * n;
+
+	switch (fp->type) {
+	case PF_USE_STDIO:
+		return fwrite(ptr, size, n, fp->fp);
+	case PF_USE_MEMORY:
+		if (bytes > fp->len) {
+			errno = ENOSPC;
+			return 0;
+		}
+
+		memcpy(fp->data, ptr, bytes);
+		fp->data += bytes;
+		fp->len -= bytes;
+		return n;
+	case PF_LEN:
+		fp->len += bytes;
+		return n;
+	default:
+		return 0;
+	}
+	return 0;
+}
+
 /*
  * Read a new set of configuration data from 
  * a policy database binary representation file.
@@ -951,12 +1005,13 @@ int hidden sepol_load_policy(void *data, size_t len)
 	convert_context_args_t args;
 	uint32_t seqno;
 	int rc = 0;
-	struct policy_file file = {
-		.type = PF_USE_MEMORY,
-		.data = data,
-		.len = len,
-		.fp = NULL
-	}, *fp = &file;
+	struct policy_file file, *fp;
+
+	policy_file_init(&file);
+	file.type = PF_USE_MEMORY;
+	file.data = data;
+	file.len = len;
+	fp = &file;
 
 	if (policydb_init(&newpolicydb))
 		return -ENOMEM;
