@@ -237,7 +237,7 @@ int sepol_user_modify(sepol_handle_t * handle,
 
 		tmp_ptr = realloc(policydb->sym_val_to_name[SYM_USERS],
 				  (policydb->p_users.nprim +
-				   1) * sizeof(user_datum_t *));
+				   1) * sizeof(char *));
 		if (!tmp_ptr)
 			goto omem;
 		policydb->sym_val_to_name[SYM_USERS] = tmp_ptr;
@@ -260,7 +260,7 @@ int sepol_user_modify(sepol_handle_t * handle,
 
 		/* Expand roles */
 		if (role_set_expand
-		    (&usrdatum->roles, &usrdatum->cache, policydb)) {
+		    (&usrdatum->roles, &usrdatum->cache, policydb, NULL)) {
 			ERR(handle, "unable to expand role set");
 			goto err;
 		}
