@@ -229,9 +229,9 @@ static int avtab_write_item(policydb_t * p,
 
 static inline void avtab_reset_merged(avtab_t * a)
 {
-	int i;
+	unsigned int i;
 	avtab_ptr_t cur;
-	for (i = 0; i < AVTAB_SIZE; i++) {
+	for (i = 0; i < a->nslot; i++) {
 		for (cur = a->htable[i]; cur; cur = cur->next)
 			cur->merged = 0;
 	}
@@ -239,7 +239,8 @@ static inline void avtab_reset_merged(avtab_t * a)
 
 static int avtab_write(struct policydb *p, avtab_t * a, struct policy_file *fp)
 {
-	int i, rc;
+	unsigned int i;
+	int rc;
 	avtab_t expa;
 	avtab_ptr_t cur;
 	uint32_t nel;
@@ -269,7 +270,7 @@ static int avtab_write(struct policydb *p, avtab_t * a, struct policy_file *fp)
 			return POLICYDB_ERROR;
 	}
 
-	for (i = 0; i < AVTAB_SIZE; i++) {
+	for (i = 0; i < a->nslot; i++) {
 		for (cur = a->htable[i]; cur; cur = cur->next) {
 			/* If old format, compute final nel.
 			   If new format, write out the items. */
@@ -290,7 +291,7 @@ static int avtab_write(struct policydb *p, avtab_t * a, struct policy_file *fp)
 			goto out;
 		}
 		avtab_reset_merged(a);
-		for (i = 0; i < AVTAB_SIZE; i++) {
+		for (i = 0; i < a->nslot; i++) {
 			for (cur = a->htable[i]; cur; cur = cur->next) {
 				if (avtab_write_item(p, cur, fp, 1, 1, NULL)) {
 					rc = -1;
@@ -958,6 +959,12 @@ static int type_write(hashtab_key_t key, hashtab_datum_t datum, void *ptr)
 	buf[items++] = cpu_to_le32(typdatum->primary);
 	if (p->policy_type != POLICY_KERN) {
 		buf[items++] = cpu_to_le32(typdatum->flavor);
+		if (p->policyvers >= MOD_POLICYDB_VERSION_PERMISSIVE)
+			buf[items++] = cpu_to_le32(typdatum->flags);
+		else if (typdatum->flags & TYPE_FLAGS_PERMISSIVE)
+			WARN(fp->handle, "Warning! Module policy version %d cannnot "
+			     "support permissive types, but one was defined",
+			     p->policyvers);
 	}
 	items2 = put_entry(buf, sizeof(uint32_t), items, fp);
 	if (items != items2)
@@ -1090,8 +1097,8 @@ static int ocontext_write(struct policydb_compat_info *info, policydb_t * p,
 					return POLICYDB_ERROR;
 				break;
 			case OCON_NODE:
-				buf[0] = cpu_to_le32(c->u.node.addr);
-				buf[1] = cpu_to_le32(c->u.node.mask);
+				buf[0] = c->u.node.addr; /* network order */
+				buf[1] = c->u.node.mask; /* network order */
 				items = put_entry(buf, sizeof(uint32_t), 2, fp);
 				if (items != 2)
 					return POLICYDB_ERROR;
@@ -1113,11 +1120,9 @@ static int ocontext_write(struct policydb_compat_info *info, policydb_t * p,
 				break;
 			case OCON_NODE6:
 				for (j = 0; j < 4; j++)
-					buf[j] =
-					    cpu_to_le32(c->u.node6.addr[j]);
+					buf[j] = c->u.node6.addr[j]; /* network order */
 				for (j = 0; j < 4; j++)
-					buf[j + 4] =
-					    cpu_to_le32(c->u.node6.mask[j]);
+					buf[j + 4] = c->u.node6.mask[j]; /* network order */
 				items = put_entry(buf, sizeof(uint32_t), 8, fp);
 				if (items != 8)
 					return POLICYDB_ERROR;
@@ -1531,8 +1536,21 @@ int policydb_write(policydb_t * p, struct policy_file *fp)
 	pd.p = p;
 
 	config = 0;
-	if (p->mls)
+	if (p->mls) {
+		if ((p->policyvers < POLICYDB_VERSION_MLS &&
+		    p->policy_type == POLICY_KERN) ||
+		    (p->policyvers < MOD_POLICYDB_VERSION_MLS &&
+		    p->policy_type == POLICY_BASE) ||
+		    (p->policyvers < MOD_POLICYDB_VERSION_MLS &&
+		    p->policy_type == POLICY_MOD)) {
+			ERR(fp->handle, "policy version %d cannot support MLS",
+			    p->policyvers);
+			return POLICYDB_ERROR;
+		}
 		config |= POLICYDB_CONFIG_MLS;
+	}
+
+	config |= (POLICYDB_CONFIG_UNKNOWN_MASK & p->handle_unknown);
 
 	/* Write the magic number and string identifiers. */
 	items = 0;
@@ -1593,6 +1611,38 @@ int policydb_write(policydb_t * p, struct policy_file *fp)
 		if (items != len)
 			return POLICYDB_ERROR;
 	}
+
+	if ((p->policyvers >= POLICYDB_VERSION_POLCAP &&
+	     p->policy_type == POLICY_KERN) ||
+	    (p->policyvers >= MOD_POLICYDB_VERSION_POLCAP &&
+	     p->policy_type == POLICY_BASE) ||
+	    (p->policyvers >= MOD_POLICYDB_VERSION_POLCAP &&
+	     p->policy_type == POLICY_MOD)) {
+		if (ebitmap_write(&p->policycaps, fp) == -1)
+			return POLICYDB_ERROR;
+	}
+
+	if (p->policyvers < POLICYDB_VERSION_PERMISSIVE &&
+	    p->policy_type == POLICY_KERN) {
+		ebitmap_node_t *tnode;
+		unsigned int i;
+
+		ebitmap_for_each_bit(&p->permissive_map, tnode, i) {
+			if (ebitmap_node_get_bit(tnode, i)) {
+				WARN(fp->handle, "Warning! Policy version %d cannot "
+				     "support permissive types, but some were defined",
+				     p->policyvers);
+				break;
+			}
+		}
+	}
+
+	if (p->policyvers >= POLICYDB_VERSION_PERMISSIVE &&
+	    p->policy_type == POLICY_KERN) {
+		if (ebitmap_write(&p->permissive_map, fp) == -1)
+			return POLICYDB_ERROR;
+	}
+
 	num_syms = info->sym_num;
 	for (i = 0; i < num_syms; i++) {
 		buf[0] = cpu_to_le32(p->symtab[i].nprim);
